#include "Document.h"
#include <nlp_string.h>
#include <nlp_config.h>
#include <nlp_filesystem.h>
#include <nlp_macros.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <assert.h>
#include "Phrase.h"
#include "Model.h"


Document_dq_t	Document::sdq_Documents;
unsigned long	Document::sul_TotalTokenCount = 0;
unsigned int	Document::sui_Types = 0;
unsigned int	Document::sui_UniquePhrases = 0;


// --------------------------------
Document::Document (void)
{
	ui_DocumentTokens = 0;
	pui_TokenIndex = NULL;
	pui_TokenType = NULL;
	pui_TokenTopic = NULL;
	pc_TokenTopicSource = NULL;
	pul_PhraseTopicTokenCounts = NULL;

	ui_DocumentPhrases = 0;
	pui_PhraseAnnotation = NULL;
	pd_TopicAnnotation = NULL;

	pd_ThetaD = NULL;
	d_R = 0;

	pd_SumThetaD = NULL;
	b_TestDocument = false;
}



// --------------------------------
Document::~Document (void)
{
	ui_DocumentTokens = 0xFF;
	if (NULL != pui_TokenIndex)
		delete[] pui_TokenIndex;
	pui_TokenIndex = NULL;
	if (NULL != pui_TokenType)
		delete[] pui_TokenType;
	pui_TokenType = NULL;
	if (NULL != pui_TokenTopic)
		delete[] pui_TokenTopic;
	pui_TokenTopic = NULL;
	if (NULL != pc_TokenTopicSource)
		delete[] pc_TokenTopicSource;
	pc_TokenTopicSource = NULL;
	if (NULL != pul_PhraseTopicTokenCounts)
		delete[] pul_PhraseTopicTokenCounts;
	pul_PhraseTopicTokenCounts = NULL;

	ui_DocumentPhrases = 0xFF;
	if (NULL != pui_PhraseAnnotation)
		delete[] pui_PhraseAnnotation;
	pui_PhraseAnnotation = NULL;
	if (NULL != pd_TopicAnnotation)
		delete[] pd_TopicAnnotation;
	pd_TopicAnnotation = NULL;

	if (NULL != pd_ThetaD)
		delete[]  pd_ThetaD;
	pd_ThetaD = (double*)0x1;
}


// --------------------------------
bool Document::Load (void)
{
	if (false == LoadTraining ())
		return false;
	if (true == (bool) (config)"use_test_document_annotations")
		return LoadTestWithAnnotations ();
	else
		return LoadTestWithoutAnnotations ();
}


// --------------------------------
bool Document::LoadTraining (void)
{
	if (0 == Phrase::sui_PhraseCount)
	{
		cerr << "[ERROR]  Phrases need to be loaded before documents." << endl;
		return false;
	}
	Document::sui_UniquePhrases = Phrase::sui_PhraseCount;

	CsvFile filePhrases ((config)"phrases_file");
	CsvFile fileWords ((config)"words_file");

	unsigned long ulPhraseCount = filePhrases.GetFirstLineValueCount ();
	unsigned long ulWordCount = fileWords.GetFirstLineValueCount ();
	Document::sui_Types = ulWordCount;
	if (ulPhraseCount != Document::sui_UniquePhrases)
	{
		cerr << "[ERROR]  Number of values in the phrases file ("
			 << (config)"phrases_file" 
			 << " is not equal to the number of unique phrases ("
			 << Document::sui_UniquePhrases << endl;
		return false;
	}
	if ((0 == ulPhraseCount) || (0 == ulWordCount))
	{
		cerr << "[ERROR]  The phrases file (" << (config)"phrases_file" 
			 << ") and the words file (" << (config)"words_file"
			 << ") should have the same number of lines. This does "
			    "not seem to be the case. Please verify."
			 << endl;
		return false;
	}

	unsigned int uiDocumentCount = 0;
	int iPhraseVector [ulPhraseCount];
	int iWordVector [ulWordCount];
	bool bPhrasesOK = filePhrases.ReadLine (iPhraseVector, &ulPhraseCount);
	bool bWordsOK = fileWords.ReadLine (iWordVector, &ulWordCount);
	bool bIgnoreUnAnnotated = ((config)"ignore_unannotated" == "true");

	while ((true == bPhrasesOK) && (true == bWordsOK))
	{
		if ((0 == ulPhraseCount) && (0 == ulWordCount))
			continue;

		bool bAnnotated = false;
		for (unsigned long i = 0; i < ulPhraseCount; ++ i)
		{
			if (0 != iPhraseVector [i])
			{
				bAnnotated = true;
				break;
			}
		}
		if ((true == bIgnoreUnAnnotated) && (false == bAnnotated))
		{
			bPhrasesOK = filePhrases.ReadLine (iPhraseVector, &ulPhraseCount);
			bWordsOK = fileWords.ReadLine (iWordVector, &ulWordCount);
			continue;
		}

		++ uiDocumentCount;
		Document* pDocument = new Document;
		sdq_Documents.push_back (pDocument);

		// process document phrase labels ...	
		for (unsigned long i = 0; i < ulPhraseCount; ++ i)
		{
			if (0 == iPhraseVector [i])
				continue;
			++ pDocument->ui_DocumentPhrases;
		}
		pDocument->pui_PhraseAnnotation = new unsigned int [pDocument->ui_DocumentPhrases];
		pDocument->pp_PhraseAnnotation = new Phrase* [pDocument->ui_DocumentPhrases];

		unsigned int j = 0;
		for (unsigned int i = 0; i < ulPhraseCount; ++ i)
		{
			if (0 == iPhraseVector [i])
				continue;
			pDocument->pui_PhraseAnnotation [j] = i;
			pDocument->pp_PhraseAnnotation [j] = Phrase::sdq_Phrases [i];
			++ j;
		}
		assert (j == pDocument->ui_DocumentPhrases);


		// process document words ...			
		if (ulWordCount != sui_Types)
		{
			cerr << "[ERROR]  All documents should be based on the same word types. "
				 << "previous documents use a set of " << sui_Types << " words, but "
				 << "document " << uiDocumentCount << " uses " << ulWordCount
				 << " words." << endl;
		}

		for (unsigned long i = 0; i < ulWordCount; ++ i)
		{
			if (0 == iWordVector [i])
				continue;
			pDocument->ui_DocumentTokens += iWordVector [i];
		}
		pDocument->pui_TokenIndex = new unsigned int [pDocument->ui_DocumentTokens];
		pDocument->pui_TokenType = new unsigned int [pDocument->ui_DocumentTokens];
		for (unsigned int i = 0; i < pDocument->ui_DocumentTokens; ++ i)
			pDocument->pui_TokenIndex [i] = i;
		unsigned int k = 0;
		for (unsigned long i = 0; i < ulWordCount; ++ i)
		{
			unsigned int uiTypeCount = iWordVector [i];
			for (unsigned int j = 0; j < uiTypeCount; ++ j)
				pDocument->pui_TokenType [k ++] = i;
		}
		assert (k == pDocument->ui_DocumentTokens);

		bPhrasesOK = filePhrases.ReadLine (iPhraseVector, &ulPhraseCount);
		bWordsOK = fileWords.ReadLine (iWordVector, &ulWordCount);
	}

	cout << "training documents loaded : " << Document::sdq_Documents.size () << endl;
	return true;
}


// --------------------------------
bool Document::LoadTestWithAnnotations (void)
{
	CsvFile filePhrases ((config)"test_phrases_file");
	CsvFile fileWords ((config)"test_words_file");

	unsigned long ulPhraseCount = filePhrases.GetFirstLineValueCount ();
	unsigned long ulWordCount = fileWords.GetFirstLineValueCount ();
	if (ulPhraseCount != Document::sui_UniquePhrases)
	{
		cerr << "[ERROR]  Number of values in the phrases file ("
			 << (config)"phrases_file" 
			 << " is not equal to the number of unique phrases ("
			 << Document::sui_UniquePhrases << endl;
		return false;
	}
	if (ulWordCount != Document::sui_Types)
	{
		cerr << "[ERROR]  Number of types in the test words file "
			 << (config)"test_words_file (" 
			 << ulWordCount
			 << ") is not equal to the number in the training words file ("
			 << Document::sui_UniquePhrases << endl;
		return false;
	}

	unsigned int uiDocumentCount = 0;
	int iPhraseVector [ulPhraseCount];
	int iWordVector [ulWordCount];
	bool bPhrasesOK = filePhrases.ReadLine (iPhraseVector, &ulPhraseCount);
	bool bWordsOK = fileWords.ReadLine (iWordVector, &ulWordCount);

	while ((true == bPhrasesOK) && (true == bWordsOK))
	{
		if ((0 == ulPhraseCount) && (0 == ulWordCount))
			continue;

		++ uiDocumentCount;
		Document* pDocument = new Document;
		pDocument->b_TestDocument = true;
		sdq_Documents.push_back (pDocument);


		// process document phrase labels ...	
		for (unsigned long i = 0; i < ulPhraseCount; ++ i)
		{
			if (0 == iPhraseVector [i])
				continue;
			++ pDocument->ui_DocumentPhrases;
		}
		pDocument->pui_PhraseAnnotation = new unsigned int [pDocument->ui_DocumentPhrases];
		pDocument->pp_PhraseAnnotation = new Phrase* [pDocument->ui_DocumentPhrases];

		unsigned int j = 0;
		for (unsigned int i = 0; i < ulPhraseCount; ++ i)
		{
			if (0 == iPhraseVector [i])
				continue;
			pDocument->pui_PhraseAnnotation [j] = i;
			pDocument->pp_PhraseAnnotation [j] = Phrase::sdq_Phrases [i];
			++ j;
		}
		assert (j == pDocument->ui_DocumentPhrases);


		// process document words ...			
		if (ulWordCount != sui_Types)
		{
			cerr << "[ERROR]  All documents should be based on the same word types. "
				 << "previous documents use a set of " << sui_Types << " words, but "
				 << "document " << uiDocumentCount << " uses " << ulWordCount
				 << " words." << endl;
		}

		for (unsigned long i = 0; i < ulWordCount; ++ i)
		{
			if (0 == iWordVector [i])
				continue;
			pDocument->ui_DocumentTokens += iWordVector [i];
		}
		pDocument->pui_TokenIndex = new unsigned int [pDocument->ui_DocumentTokens];
		pDocument->pui_TokenType = new unsigned int [pDocument->ui_DocumentTokens];
		for (unsigned int i = 0; i < pDocument->ui_DocumentTokens; ++ i)
			pDocument->pui_TokenIndex [i] = i;
		unsigned int k = 0;
		for (unsigned long i = 0; i < ulWordCount; ++ i)
		{
			unsigned int uiTypeCount = iWordVector [i];
			for (unsigned int j = 0; j < uiTypeCount; ++ j)
				pDocument->pui_TokenType [k ++] = i;
		}
		assert (k == pDocument->ui_DocumentTokens);

		bWordsOK = fileWords.ReadLine (iWordVector, &ulWordCount);
	}

	cout << "test document laoded with annotations" << endl;
	cout << "test documents loaded  : " << uiDocumentCount << endl;
	cout << "total documents loaded : " << Document::sdq_Documents.size () << endl;
	return true;
}


// --------------------------------
bool Document::LoadTestWithoutAnnotations (void)
{
	CsvFile fileWords ((config)"test_words_file");

	unsigned long ulWordCount = fileWords.GetFirstLineValueCount ();
	if (ulWordCount != Document::sui_Types)
	{
		cerr << "[ERROR]  Number of types in the test words file "
			 << (config)"test_words_file (" 
			 << ulWordCount
			 << ") is not equal to the number in the training words file ("
			 << Document::sui_UniquePhrases << endl;
		return false;
	}

	unsigned int uiDocumentCount = 0;
	int iWordVector [ulWordCount];
	bool bWordsOK = fileWords.ReadLine (iWordVector, &ulWordCount);

	while (true == bWordsOK)
	{
		if (0 == ulWordCount)
			continue;

		++ uiDocumentCount;
		Document* pDocument = new Document;
		pDocument->b_TestDocument = true;
		sdq_Documents.push_back (pDocument);


		// process document words ...			
		if (ulWordCount != sui_Types)
		{
			cerr << "[ERROR]  All documents should be based on the same word types. "
				 << "previous documents use a set of " << sui_Types << " words, but "
				 << "document " << uiDocumentCount << " uses " << ulWordCount
				 << " words." << endl;
		}

		for (unsigned long i = 0; i < ulWordCount; ++ i)
		{
			if (0 == iWordVector [i])
				continue;
			pDocument->ui_DocumentTokens += iWordVector [i];
		}
		pDocument->pui_TokenIndex = new unsigned int [pDocument->ui_DocumentTokens];
		pDocument->pui_TokenType = new unsigned int [pDocument->ui_DocumentTokens];
		for (unsigned int i = 0; i < pDocument->ui_DocumentTokens; ++ i)
			pDocument->pui_TokenIndex [i] = i;
		unsigned int k = 0;
		for (unsigned long i = 0; i < ulWordCount; ++ i)
		{
			unsigned int uiTypeCount = iWordVector [i];
			for (unsigned int j = 0; j < uiTypeCount; ++ j)
				pDocument->pui_TokenType [k ++] = i;
		}
		assert (k == pDocument->ui_DocumentTokens);

		bWordsOK = fileWords.ReadLine (iWordVector, &ulWordCount);
	}

	cout << "test documents loaded  : " << uiDocumentCount << endl;
	cout << "total documents loaded : " << Document::sdq_Documents.size () << endl;
	return true;
}



// --------------------------------
void Document::InitParameters (void)
{
	// initialize random number generator				
	gsl_rng_env_setup ();
	gsl_rng* pGslRng = gsl_rng_alloc (gsl_rng_default);

	// initialize discrete random variable GSL structs	
	double* pdTopicProb = new double [Model::sui_Topics];
	for (unsigned int i = 0; i < Model::sui_Topics; ++ i)
		pdTopicProb [i] = 1;
	gsl_ran_discrete_t* pGslTopicProb = 
		gsl_ran_discrete_preproc (Model::sui_Topics, pdTopicProb);
	delete[] pdTopicProb;
	double dTopicSourceProb [2] = {1, 1};
	gsl_ran_discrete_t* pGslTopicSourceProb =
		gsl_ran_discrete_preproc (2, dTopicSourceProb);

	ITERATE (Document_dq_t, sdq_Documents, ite)
	{
		Document* pDocument = *ite;

		pDocument->pd_ThetaD = new double [Model::sui_Topics];
		for (unsigned int i = 0; i < Model::sui_Topics; ++ i)
			pDocument->pd_ThetaD [i] = 1;
		pDocument->d_R = 0.1;

		if (true == pDocument->b_TestDocument)
		{
			pDocument->pd_SumThetaD = new double [Model::sui_Topics];
			memset (pDocument->pd_SumThetaD, 0, sizeof (double) * Model::sui_Topics);
		}

		// the topics of the label phrases is initialized for the	
		// phrase clustering later, just allocation memory here ...	
		pDocument->pd_TopicAnnotation = new double [Model::sui_Topics];
		memset (pDocument->pd_TopicAnnotation, 0, sizeof (double) * Model::sui_Topics);

		pDocument->pul_PhraseTopicTokenCounts = new unsigned long [Model::sui_Topics];
		memset (pDocument->pul_PhraseTopicTokenCounts, 0, sizeof (unsigned long) * Model::sui_Topics);

		// randomly initialize the topics of the text tokens and	
		// the token topic sources ...								
		pDocument->pui_TokenTopic = new unsigned int [pDocument->ui_DocumentTokens];
		pDocument->pc_TokenTopicSource = new char [pDocument->ui_DocumentTokens];
		for (unsigned int i = 0; i < pDocument->ui_DocumentTokens; ++ i)
		{
			pDocument->pui_TokenTopic [i] = 
				(unsigned int) gsl_ran_discrete (pGslRng, pGslTopicProb);
			pDocument->pc_TokenTopicSource [i] =
				(char) gsl_ran_discrete (pGslRng, pGslTopicSourceProb);
		}
	}

	gsl_ran_discrete_free (pGslTopicProb);
	gsl_ran_discrete_free (pGslTopicSourceProb);
	gsl_rng_free (pGslRng);
}


// --------------------------------
void Document::Clear (void)
{
	ITERATE (Document_dq_t, sdq_Documents, ite)
		delete *ite;
	sdq_Documents.clear ();
}


// --------------------------------
void Document::SaveTestDocumentTopics (void)
{
	{
		String sFileName;
		sFileName << (config)"test_document_topic_prediction_file" << ".avg.csv";
		CsvFile file (sFileName, ios_base::out);

		ITERATE (Document_dq_t, sdq_Documents, ite)
		{
			Document* pDocument = *ite;
			if (false == pDocument->b_TestDocument)
				continue;
			for (unsigned int t = 0; t < Model::sui_Topics; ++ t)
				pDocument->pd_SumThetaD [t] /= (double)Model::sui_TopicSamples;

			file.WriteLine (pDocument->pd_SumThetaD, Model::sui_Topics);
		}

		file.flush ();
		file.close ();
	}
	{
		String sFileName;
		sFileName << (config)"test_document_topic_prediction_file" << ".cut.csv";
		CsvFile file (sFileName, ios_base::out);

		ITERATE (Document_dq_t, sdq_Documents, ite)
		{
			Document* pDocument = *ite;
			if (false == pDocument->b_TestDocument)
				continue;
			file.WriteLine (pDocument->pd_ThetaD, Model::sui_Topics);
		}

		file.flush ();
		file.close ();
	}
}



