#ifndef __MODEL__
#define __MODEL__

#include <gsl/gsl_rng.h>
#include <pthread.h>
#include "Document.h"


class ThreadInfo
{
	public:
		unsigned int	ui_Topics;
		unsigned int	ui_Types;

		pthread_t		o_Thread;
		pthread_mutex_t	mtx_Start;
		pthread_mutex_t	mtx_Done;
		long			l_DocumentsForSampling;
		Document**		pp_DocumentsForSampling;
		Document_dq_t	dq_Documents;

		gsl_rng*		p_GslRng;
		double			d_AlphaD;
		double**		ppd_Phi;
		double*			pd_TypeTopicCountData;
		bool			b_Done;

		ThreadInfo (void);
		~ThreadInfo (void);
};


class Model
{
	private:
		int			i_ThreadCount;
		ThreadInfo* p_ThreadInfo;
		Phrase**	pp_PhrasesForSampling;

	public:
		static unsigned int sui_Topics;
		static unsigned int sui_TopicSamples;
		static bool			sb_PredictTestDocumentProperties;

		static double	d_TopicSmoothing;
		double			d_Alpha;
		double*			pd_Theta;

		double			d_AlphaD;
		double			d_Beta;
		double*			pd_PhiData;
		double**		ppd_Phi;
		unsigned int*	pui_TopicLogData;

		double			d_TopicSmoothingScaling;
		double			d_AlphaScaling;
		double			d_AlphaDScaling;
		double			d_BetaScaling;

		gsl_rng*		p_GslRng;


		Model (void);
		~Model (void);

		void Init (void);
		void InitThreads (void);
		void InitParameters (void);

		void Sample (void);
		void SampleClusters (bool _bLinkToDocuments = false);
		void SampleLanguageModels (void);
		static void* SampleDocumentLanguageModels (void* _pParam);

		void SaveClusterParams (void);
		void SaveLanguageModelParams (int _iIteration);
		void LoadClusterParams (void);
		void LoadLanguageModelParams (int _iIteration);

		void SaveClusterState (void);
		void SaveLanguageModelState (void);
		void LoadClusterState (void);
		void LoadLanguageModelState (void);
};


#endif
