#include <nlp_string.h>
#include <nlp_config.h>
#include <nlp_filesystem.h>
#include <nlp_macros.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include "Phrase.h"
#include "Model.h"


unsigned int	Phrase::sui_PhraseCount = 0;
Phrase_dq_t		Phrase::sdq_Phrases;


// ------------------------------------------------------
Phrase::Phrase (void)
{
	pd_SimilarityVector = NULL;
	pd_TopicProbability = NULL;
	ui_Topic = (int)0xFF;
}

Phrase::~Phrase (void)
{
	if (NULL != pd_SimilarityVector)
		delete[] pd_SimilarityVector;
	pd_SimilarityVector = (double*)0x1;
	if (NULL != pd_TopicProbability)
		delete[] pd_TopicProbability;
	pd_TopicProbability = (double*)0x1;
	ui_Topic = (int)0xFE;
}


// ------------------------------------------------------
bool Phrase::Load (void)
{
	CsvFile fileLex ((config)"lex_file");
	CsvFile fileCooc ((config)"cooc_file");

	unsigned int uiPhraseCount = 0;
	sui_PhraseCount = fileLex.GetFirstLineValueCount ();
	if (sui_PhraseCount != fileCooc.GetFirstLineValueCount ())
	{
		cerr << "[ERROR]  The lexical & co-occurrence vectors need to be of "
			 << "the same length. This doesn't seem to be the case. Please check files "
			 << (config)"lex_file" << " and " << (config)"cooc_file" << endl;
		return false;
	}

	double dLexVector [sui_PhraseCount];
	double dCoocVector [sui_PhraseCount];

	unsigned long ulPhraseCount = sui_PhraseCount;
	bool bLexOK = fileLex.ReadLine (dLexVector, &ulPhraseCount);
	bool bCoocOK = fileCooc.ReadLine (dCoocVector, &ulPhraseCount);
	while ((true == bCoocOK) && (true == bLexOK))
	{
		++ uiPhraseCount;
		Phrase* pPhrase = new Phrase;
		pPhrase->pd_TopicProbability = new double [Model::sui_Topics];
		sdq_Phrases.push_back (pPhrase);

		pPhrase->pd_SimilarityVector = new double [sui_PhraseCount];
		for (unsigned int i = 0; i < sui_PhraseCount; ++ i)
		{
			double dSimilarity = 0.5 * (double)dCoocVector [i] + 0.5 * (double)dLexVector [i];
			if (dSimilarity > 0.9)
				dSimilarity = 0.9;
			if (dSimilarity < 0.1)
				dSimilarity = 0.1;
			pPhrase->pd_SimilarityVector [i] = dSimilarity;
		}

		bLexOK = fileLex.ReadLine (dLexVector, &ulPhraseCount);
		bCoocOK = fileCooc.ReadLine (dCoocVector, &ulPhraseCount);
	}
	return true;
}


// ------------------------------------------------------
void Phrase::Clear (void)
{
	ITERATE (Phrase_dq_t, sdq_Phrases, ite)
		delete *ite;
	sdq_Phrases.clear ();
}


// ------------------------------------------------------
void Phrase::InitParameters (void)
{
	// initialize random number generator				
	gsl_rng_env_setup ();
	gsl_rng* pGslRng = gsl_rng_alloc (gsl_rng_default);

	// initialize discrete random variable GSL structs	
	double* pdTopicProb = new double [Model::sui_Topics];
	for (unsigned int i = 0; i < Model::sui_Topics; ++ i)
		pdTopicProb [i] = 1;
	gsl_ran_discrete_t* pGslTopicProb = 
		gsl_ran_discrete_preproc (Model::sui_Topics, pdTopicProb);

	ITERATE (Phrase_dq_t, sdq_Phrases, ite)
	{
		Phrase* pPhrase = *ite;
		pPhrase->ui_Topic =
			(unsigned int) gsl_ran_discrete (pGslRng, pGslTopicProb);
	}

	gsl_ran_discrete_free (pGslTopicProb);
	gsl_rng_free (pGslRng);
}



