#ifndef __NLP_FILE_SYSTEM__
#define __NLP_FILE_SYSTEM__

#include <iostream>
#include <fstream>
#include <stdio.h>
#include <ext/stdio_filebuf.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <nlp_string.h>
#include <deque>
using namespace std;


typedef deque <double>	double_dq_t;
typedef deque <float>	float_dq_t;
typedef deque <long>	long_dq_t;
typedef deque <int>		int_dq_t;


// ------------------------------------------------------
class Path
{
	public:
		static bool Exists (const char* _zPath);
		static bool Readable (const char* _zPath);
		static bool Writable (const char* _zPath);
		static bool Runable (const char* _zPath);
		static bool Access (const char* _zPath, int _iMode);

		static bool CreatePath (const char* _zPath, mode_t _iMode);
		static bool RemovePath (const char* _zPath);
		static bool MovePath (const char* _zPath);

		static bool MoveFile (const char* _zFileName);
		static bool RemoveFile (const char* _zFileName);

		static bool GetFileList (const char* _zPath, String_dq_t& _rdqFileNames);
		static bool GetPathList (const char* _zPath, String_dq_t& _rdqFileNames);

		static String GetCanonicalPath (const char* _zPath);
};


// ------------------------------------------------------
class File : public fstream
{
	protected:
		enum FileStatus_e
		{
			en_fs_initial,
			en_fs_open,
			en_fs_open_failed,
			en_fs_closed,
			en_fs_destroyed
		};

		FILE*							p_File;
		__gnu_cxx::stdio_filebuf<char>*	p_FDBuf;
		String							s_FileName;
		std::ios_base::openmode			e_Mode;
		FileStatus_e					e_FileStatus;

		// the following two lines are a hack - for some reason, gcc3.3.5 doesn't disambiguate	
		// getline to the std::string function correctly when it is used inside a class			
		// inherited from fstream.  These static inlines bypass this problem.					
		// static void getline (fstream& _rfsFile, String& _rLine) {std::getline (_rfsFile, _rLine);};
		// static void getline (ifstream& _rfsFile, String& _rLine) {std::getline (_rfsFile, _rLine);};

	public:
		File (void);
		File (const char* _zFileName, std::ios_base::openmode _eMode = std::ios_base::in);
		~File (void);

		bool Open (const char* _zFileName, std::ios_base::openmode _eMode = std::ios_base::in);
		void Close (void);
		
		bool ReadLines (String_dq_t& _rdqLines);
		bool ReadLine (String& _rsLine);

		bool WriteLines (String_dq_t& _rdqLines);
		bool WriteLine (const char* _zLine);

		static bool ReadLines (const char* _zFileName, String_dq_t& _rdqLines);
		static bool ReadLines (const char* _zFileName, String_set_t& _rsetLines);
		static bool WriteLines (const char* _zFileName, String_dq_t& _rdqLines);
		static bool WriteLines (const char* _zFileName, String_set_t& _rsetLines);
};


// ------------------------------------------------------
class CsvFile : public File
{
		unsigned long ul_LineNumber;

	public:
		CsvFile (void);
		CsvFile (const char* _zFileName, std::ios_base::openmode _eMode = std::ios_base::in);
		~CsvFile (void);

		unsigned long GetFirstLineValueCount (void);
	
		bool ReadLine (double_dq_t& _dqValues, unsigned long* _pulCount);
		bool ReadLine (float_dq_t& _dqValues, unsigned long* _pulCount);
		bool ReadLine (long_dq_t& _dqValues, unsigned long* _pulCount);
		bool ReadLine (int_dq_t& _dqValues, unsigned long* _pulCount);

		bool ReadLine (double* _pValues, unsigned long* _pulCount);
		bool ReadLine (float* _pValues, unsigned long* _pulCount);
		bool ReadLine (long* _pValues, unsigned long* _pulCount);
		bool ReadLine (unsigned long* _pValues, unsigned long* _pulCount);
		bool ReadLine (int* _pValues, unsigned long* _pulCount);
		bool ReadLine (unsigned int* _pValues, unsigned long* _pulCount);
		bool ReadLine (char* _pValues, unsigned long* _pulCount);
		bool ReadLine (unsigned char* _pValues, unsigned long* _pulCount);
	
		bool WriteLine (double_dq_t& _dqValues);
		bool WriteLine (float_dq_t& _dqValues);
		bool WriteLine (long_dq_t& _dqValues);
		bool WriteLine (int_dq_t& _dqValues);

		bool WriteLine (double* _pValues, unsigned long _ulCount);
		bool WriteLine (float* _pValues, unsigned long _ulCount);
		bool WriteLine (long* _pValues, unsigned long _ulCount);
		bool WriteLine (unsigned long* _pValues, unsigned long _ulCount);
		bool WriteLine (int* _pValues, unsigned long _ulCount);
		bool WriteLine (unsigned int* _pValues, unsigned long _ulCount);
		bool WriteLine (char* _pValues, unsigned long _ulCount);
		bool WriteLine (unsigned char* _pValues, unsigned long _ulCount);
};


#endif
