#include <iostream>
#include <fstream>
#include "nlp_filesystem.h"
#include <nlp_macros.h>
using namespace std;



// -----------------------------------------------------------------
CsvFile::CsvFile (void)
	: File ()
{
	ul_LineNumber = 0;
}


// -----------------------------------------------------------------
CsvFile::CsvFile (const char* _zFileName, ios_base::openmode _eMode)
	: File (_zFileName, _eMode)
{
	ul_LineNumber = 0;
}


// -----------------------------------------------------------------
CsvFile::~CsvFile (void)
{
	ul_LineNumber = 0;
}


// -----------------------------------------------------------------
unsigned long CsvFile::GetFirstLineValueCount (void)
{
	if (en_fs_open != e_FileStatus)
	{
		cerr << "[ERROR]  Unable to read from file '" << s_FileName 
			 << "' since it has not yet been opened !" << endl;
		return false;
	}
	if (0 == (ios_base::in & e_Mode))
	{
		cerr << "[ERROR]  Unable to read from file '" << s_FileName 
			 << "' since it was not opened in read mode !" << endl;
		return false;
	}

	// remember starting location & state ...
	ios::iostate iStartingState = ios::rdstate ();
	streampos iStartingPos = istream::tellg ();

	// go to start of file & find out number of values on the line.
	istream::seekg (0, ios::beg);
	String sLine;
	std::getline (*this, sLine);
	char* pStart = (char*)(const char*)sLine;

	unsigned long ulValues = 0;
	while ('\0' != *pStart)
	{
		while ((' ' == *pStart) || ('\t' == *pStart))
			++ pStart;

		++ ulValues;
		if (',' == *pStart)
		{
			++ pStart;
			continue;
		}
			
		pStart = strchr (pStart, ',');
		if (NULL == pStart)
			break;
		++ pStart;
	}

	// reset to starting location & state ...
	istream::seekg (iStartingPos);
	clear (iStartingState);

	return ulValues;
}


// -----------------------------------------------------------------
bool CsvFile::ReadLine (double_dq_t& _dqValues, unsigned long* _pulCount)
{
	String sLine;
	if (false == File::ReadLine (sLine))
		return false;
	char* pStart = (char*)(const char*)sLine;
	
	unsigned long ulValues = 0;
	while ('\0' != *pStart)
	{
		while ((' ' == *pStart) || ('\t' == *pStart))
			++ pStart;

		++ ulValues;
		if (',' == *pStart)
		{
			_dqValues.push_back (0);
			++ pStart;
			continue;
		}
		_dqValues.push_back (atof (pStart));
		pStart = strchr (pStart, ',');
		if (NULL == pStart)
			break;
		++ pStart;
	}
	if ((0 != *_pulCount) && (ulValues != *_pulCount))
	{
		cerr << "[ERROR]  CSV file " << s_FileName << " was expected to have " 
			 << *_pulCount << " values on line " << ul_LineNumber + 1
			 << " however, the line had " << ulValues << " values." << endl;
		return false;
	}

	*_pulCount = ulValues;
	++ ul_LineNumber;
	return true;
}


// -----------------------------------------------------------------
bool CsvFile::ReadLine (float_dq_t& _dqValues, unsigned long* _pulCount)
{
	String sLine;
	if (false == File::ReadLine (sLine))
		return false;
	char* pStart = (char*)(const char*)sLine;
	
	unsigned long ulValues = 0;
	while ('\0' != *pStart)
	{
		while ((' ' == *pStart) || ('\t' == *pStart))
			++ pStart;

		if (',' == *pStart)
		{
			_dqValues.push_back (0);
			++ ulValues;
			++ pStart;
			continue;
		}
		_dqValues.push_back (atof (pStart));
		++ ulValues;
		pStart = strchr (pStart, ',');
		if (NULL == pStart)
			break;
		++ pStart;
	}
	if ((0 != *_pulCount) && (ulValues != *_pulCount))
	{
		cerr << "[ERROR]  CSV file " << s_FileName << " was expected to have " 
			 << *_pulCount << " values on line " << ul_LineNumber + 1
			 << " however, the line had " << ulValues << " values." << endl;
		return false;
	}

	*_pulCount = ulValues;
	++ ul_LineNumber;
	return true;
}


// -----------------------------------------------------------------
bool CsvFile::ReadLine (long_dq_t& _dqValues, unsigned long* _pulCount)
{
	String sLine;
	if (false == File::ReadLine (sLine))
		return false;
	char* pStart = (char*)(const char*)sLine;
	
	unsigned long ulValues = 0;
	while ('\0' != *pStart)
	{
		while ((' ' == *pStart) || ('\t' == *pStart))
			++ pStart;

		if (',' == *pStart)
		{
			_dqValues.push_back (0);
			++ ulValues;
			++ pStart;
			continue;
		}
		_dqValues.push_back (atol (pStart));
		++ ulValues;
		pStart = strchr (pStart, ',');
		if (NULL == pStart)
			break;
		++ pStart;
	}
	if ((0 != *_pulCount) && (ulValues != *_pulCount))
	{
		cerr << "[ERROR]  CSV file " << s_FileName << " was expected to have " 
			 << *_pulCount << " values on line " << ul_LineNumber + 1
			 << " however, the line had " << ulValues << " values." << endl;
		return false;
	}

	*_pulCount = ulValues;
	++ ul_LineNumber;
	return true;
}


// -----------------------------------------------------------------
bool CsvFile::ReadLine (int_dq_t& _dqValues, unsigned long* _pulCount)
{
	String sLine;
	if (false == File::ReadLine (sLine))
		return false;
	char* pStart = (char*)(const char*)sLine;
	
	unsigned long ulValues = 0;
	while ('\0' != *pStart)
	{
		while ((' ' == *pStart) || ('\t' == *pStart))
			++ pStart;

		if (',' == *pStart)
		{
			_dqValues.push_back (0);
			++ ulValues;
			++ pStart;
			continue;
		}
		_dqValues.push_back (atoi (pStart));
		++ ulValues;
		pStart = strchr (pStart, ',');
		if (NULL == pStart)
			break;
		++ pStart;
	}
	if ((0 != *_pulCount) && (ulValues != *_pulCount))
	{
		cerr << "[ERROR]  CSV file " << s_FileName << " was expected to have " 
			 << *_pulCount << " values on line " << ul_LineNumber + 1
			 << " however, the line had " << ulValues << " values." << endl;
		return false;
	}

	*_pulCount = ulValues;
	++ ul_LineNumber;
	return true;
}


// -----------------------------------------------------------------
bool CsvFile::ReadLine (double* _pValues, unsigned long* _pulCount)
{
	String sLine;
	if (false == File::ReadLine (sLine))
		return false;
	char* pStart = (char*)(const char*)sLine;
	
	unsigned long ulValues = 0;
	while ('\0' != *pStart)
	{
		while ((' ' == *pStart) || ('\t' == *pStart))
			++ pStart;

		if (',' == *pStart)
		{
			*(_pValues ++) = 0;
			++ ulValues;
			++ pStart;
			continue;
		}
		*(_pValues ++) = atof (pStart);
		++ ulValues;
		pStart = strchr (pStart, ',');
		if (NULL == pStart)
			break;
		++ pStart;
	}
	if ((0 != *_pulCount) && (ulValues != *_pulCount))
	{
		cerr << "[ERROR]  CSV file " << s_FileName << " was expected to have " 
			 << *_pulCount << " values on line " << ul_LineNumber + 1
			 << " however, the line had " << ulValues << " values." << endl;
		return false;
	}

	*_pulCount = ulValues;
	++ ul_LineNumber;
	return true;
}


// -----------------------------------------------------------------
bool CsvFile::ReadLine (float* _pValues, unsigned long* _pulCount)
{
	String sLine;
	if (false == File::ReadLine (sLine))
		return false;
	char* pStart = (char*)(const char*)sLine;
	
	unsigned long ulValues = 0;
	while ('\0' != *pStart)
	{
		while ((' ' == *pStart) || ('\t' == *pStart))
			++ pStart;

		if (',' == *pStart)
		{
			*(_pValues ++) = 0;
			++ ulValues;
			++ pStart;
			continue;
		}
		*(_pValues ++) = atof (pStart);
		++ ulValues;
		pStart = strchr (pStart, ',');
		if (NULL == pStart)
			break;
		++ pStart;
	}
	if ((0 != *_pulCount) && (ulValues != *_pulCount))
	{
		cerr << "[ERROR]  CSV file " << s_FileName << " was expected to have " 
			 << *_pulCount << " values on line " << ul_LineNumber + 1
			 << " however, the line had " << ulValues << " values." << endl;
		return false;
	}

	*_pulCount = ulValues;
	++ ul_LineNumber;
	return true;
}


// -----------------------------------------------------------------
bool CsvFile::ReadLine (long* _pValues, unsigned long* _pulCount)
{
	String sLine;
	if (false == File::ReadLine (sLine))
		return false;
	char* pStart = (char*)(const char*)sLine;
	
	unsigned long ulValues = 0;
	while ('\0' != *pStart)
	{
		while ((' ' == *pStart) || ('\t' == *pStart))
			++ pStart;

		if (',' == *pStart)
		{
			*(_pValues ++) = 0;
			++ ulValues;
			++ pStart;
			continue;
		}
		*(_pValues ++) = atol (pStart);
		++ ulValues;
		pStart = strchr (pStart, ',');
		if (NULL == pStart)
			break;
		++ pStart;
	}
	if ((0 != *_pulCount) && (ulValues != *_pulCount))
	{
		cerr << "[ERROR]  CSV file " << s_FileName << " was expected to have " 
			 << *_pulCount << " values on line " << ul_LineNumber + 1
			 << " however, the line had " << ulValues << " values." << endl;
		return false;
	}

	*_pulCount = ulValues;
	++ ul_LineNumber;
	return true;
}


// -----------------------------------------------------------------
bool CsvFile::ReadLine (unsigned long* _pValues, unsigned long* _pulCount)
{
	String sLine;
	if (false == File::ReadLine (sLine))
		return false;
	char* pStart = (char*)(const char*)sLine;
	
	unsigned long ulValues = 0;
	while ('\0' != *pStart)
	{
		while ((' ' == *pStart) || ('\t' == *pStart))
			++ pStart;

		if (',' == *pStart)
		{
			*(_pValues ++) = 0;
			++ ulValues;
			++ pStart;
			continue;
		}
		*(_pValues ++) = atol (pStart);
		++ ulValues;
		pStart = strchr (pStart, ',');
		if (NULL == pStart)
			break;
		++ pStart;
	}
	if ((0 != *_pulCount) && (ulValues != *_pulCount))
	{
		cerr << "[ERROR]  CSV file " << s_FileName << " was expected to have " 
			 << *_pulCount << " values on line " << ul_LineNumber + 1
			 << " however, the line had " << ulValues << " values." << endl;
		return false;
	}

	*_pulCount = ulValues;
	++ ul_LineNumber;
	return true;
}


// -----------------------------------------------------------------
bool CsvFile::ReadLine (int* _pValues, unsigned long* _pulCount)
{
	String sLine;
	if (false == File::ReadLine (sLine))
		return false;
	char* pStart = (char*)(const char*)sLine;
	
	unsigned long ulValues = 0;
	while ('\0' != *pStart)
	{
		while ((' ' == *pStart) || ('\t' == *pStart))
			++ pStart;

		if (',' == *pStart)
		{
			*(_pValues ++) = 0;
			++ ulValues;
			++ pStart;
			continue;
		}
		*(_pValues ++) = atoi (pStart);
		++ ulValues;
		pStart = strchr (pStart, ',');
		if (NULL == pStart)
			break;
		++ pStart;
	}
	if ((0 != *_pulCount) && (ulValues != *_pulCount))
	{
		cerr << "[ERROR]  CSV file " << s_FileName << " was expected to have " 
			 << *_pulCount << " values on line " << ul_LineNumber + 1
			 << " however, the line had " << ulValues << " values." << endl;
		return false;
	}

	*_pulCount = ulValues;
	++ ul_LineNumber;
	return true;
}


// -----------------------------------------------------------------
bool CsvFile::ReadLine (unsigned int* _pValues, unsigned long* _pulCount)
{
	String sLine;
	if (false == File::ReadLine (sLine))
		return false;
	char* pStart = (char*)(const char*)sLine;
	
	unsigned long ulValues = 0;
	while ('\0' != *pStart)
	{
		while ((' ' == *pStart) || ('\t' == *pStart))
			++ pStart;

		if (',' == *pStart)
		{
			*(_pValues ++) = 0;
			++ ulValues;
			++ pStart;
			continue;
		}
		*(_pValues ++) = atoi (pStart);
		++ ulValues;
		pStart = strchr (pStart, ',');
		if (NULL == pStart)
			break;
		++ pStart;
	}
	if ((0 != *_pulCount) && (ulValues != *_pulCount))
	{
		cerr << "[ERROR]  CSV file " << s_FileName << " was expected to have " 
			 << *_pulCount << " values on line " << ul_LineNumber + 1
			 << " however, the line had " << ulValues << " values." << endl;
		return false;
	}

	*_pulCount = ulValues;
	++ ul_LineNumber;
	return true;
}


// -----------------------------------------------------------------
bool CsvFile::ReadLine (char* _pValues, unsigned long* _pulCount)
{
	String sLine;
	if (false == File::ReadLine (sLine))
		return false;
	char* pStart = (char*)(const char*)sLine;
	
	unsigned long ulValues = 0;
	while ('\0' != *pStart)
	{
		while ((' ' == *pStart) || ('\t' == *pStart))
			++ pStart;

		if (',' == *pStart)
		{
			*(_pValues ++) = 0;
			++ ulValues;
			++ pStart;
			continue;
		}
		*(_pValues ++) = (char)atoi (pStart);
		++ ulValues;
		pStart = strchr (pStart, ',');
		if (NULL == pStart)
			break;
		++ pStart;
	}
	if ((0 != *_pulCount) && (ulValues != *_pulCount))
	{
		cerr << "[ERROR]  CSV file " << s_FileName << " was expected to have " 
			 << *_pulCount << " values on line " << ul_LineNumber + 1
			 << " however, the line had " << ulValues << " values." << endl;
		return false;
	}

	*_pulCount = ulValues;
	++ ul_LineNumber;
	return true;
}


// -----------------------------------------------------------------
bool CsvFile::ReadLine (unsigned char* _pValues, unsigned long* _pulCount)
{
	String sLine;
	if (false == File::ReadLine (sLine))
		return false;
	char* pStart = (char*)(const char*)sLine;
	
	unsigned long ulValues = 0;
	while ('\0' != *pStart)
	{
		while ((' ' == *pStart) || ('\t' == *pStart))
			++ pStart;

		if (',' == *pStart)
		{
			*(_pValues ++) = 0;
			++ ulValues;
			++ pStart;
			continue;
		}
		*(_pValues ++) = (char)atoi (pStart);
		++ ulValues;
		pStart = strchr (pStart, ',');
		if (NULL == pStart)
			break;
		++ pStart;
	}
	if ((0 != *_pulCount) && (ulValues != *_pulCount))
	{
		cerr << "[ERROR]  CSV file " << s_FileName << " was expected to have " 
			 << *_pulCount << " values on line " << ul_LineNumber + 1
			 << " however, the line had " << ulValues << " values." << endl;
		return false;
	}

	*_pulCount = ulValues;
	++ ul_LineNumber;
	return true;
}



// -----------------------------------------------------------------
bool CsvFile::WriteLine (double_dq_t& _dqValues)
{
	this->precision (15);
	this->setf (ios::scientific);
	bool bFirst = true;
	ITERATE (double_dq_t, _dqValues, ite)
	{
		if (false == bFirst)
			*this << ',';
		else
			bFirst = false;
		*this << *ite;
	}
	*this << endl;
	return true;
}


// -----------------------------------------------------------------
bool CsvFile::WriteLine (float_dq_t& _dqValues)
{
	this->precision (15);
	this->setf (ios::scientific);
	bool bFirst = true;
	ITERATE (float_dq_t, _dqValues, ite)
	{
		if (false == bFirst)
			*this << ',';
		else
			bFirst = false;
		*this << *ite;
	}
	*this << endl;
	return true;
}


// -----------------------------------------------------------------
bool CsvFile::WriteLine (long_dq_t& _dqValues)
{
	bool bFirst = true;
	ITERATE (long_dq_t, _dqValues, ite)
	{
		if (false == bFirst)
			*this << ',';
		else
			bFirst = false;
		*this << *ite;
	}
	*this << endl;
	return true;
}

// -----------------------------------------------------------------
bool CsvFile::WriteLine (int_dq_t& _dqValues)
{
	bool bFirst = true;
	ITERATE (int_dq_t, _dqValues, ite)
	{
		if (false == bFirst)
			*this << ',';
		else
			bFirst = false;
		*this << *ite;
	}
	*this << endl;
	return true;
}


// -----------------------------------------------------------------
bool CsvFile::WriteLine (double* _pValues, unsigned long _ulCount)
{
	if (0 == _ulCount)
		return true;
	this->precision (15);
	this->setf (ios::scientific);
	*this << _pValues [0];
	for (unsigned long i = 1; i < _ulCount; ++ i)
		*this << ',' << _pValues [i];
	*this << endl;
	return true;
}


// -----------------------------------------------------------------
bool CsvFile::WriteLine (float* _pValues, unsigned long _ulCount)
{
	if (0 == _ulCount)
		return true;
	this->precision (15);
	this->setf (ios::scientific);
	*this << _pValues [0];
	for (unsigned long i = 1; i < _ulCount; ++ i)
		*this << ',' << _pValues [i];
	*this << endl;
	return true;
}


// -----------------------------------------------------------------
bool CsvFile::WriteLine (long* _pValues, unsigned long _ulCount)
{
	if (0 == _ulCount)
		return true;
	*this << _pValues [0];
	for (unsigned long i = 1; i < _ulCount; ++ i)
		*this << ',' << _pValues [i];
	*this << endl;
	return true;
}


// -----------------------------------------------------------------
bool CsvFile::WriteLine (unsigned long* _pValues, unsigned long _ulCount)
{
	if (0 == _ulCount)
		return true;
	*this << _pValues [0];
	for (unsigned long i = 1; i < _ulCount; ++ i)
		*this << ',' << _pValues [i];
	*this << endl;
	return true;
}


// -----------------------------------------------------------------
bool CsvFile::WriteLine (int* _pValues, unsigned long _ulCount)
{
	if (0 == _ulCount)
		return true;
	*this << _pValues [0];
	for (unsigned long i = 1; i < _ulCount; ++ i)
		*this << ',' << _pValues [i];
	*this << endl;
	return true;
}


// -----------------------------------------------------------------
bool CsvFile::WriteLine (unsigned int* _pValues, unsigned long _ulCount)
{
	if (0 == _ulCount)
		return true;
	*this << _pValues [0];
	for (unsigned long i = 1; i < _ulCount; ++ i)
		*this << ',' << _pValues [i];
	*this << endl;
	return true;
}


// -----------------------------------------------------------------
bool CsvFile::WriteLine (char* _pValues, unsigned long _ulCount)
{
	if (0 == _ulCount)
		return true;
	*this << (unsigned int)_pValues [0];
	for (unsigned long i = 1; i < _ulCount; ++ i)
		*this << ',' << (unsigned int)_pValues [i];
	*this << endl;
	return true;
}


// -----------------------------------------------------------------
bool CsvFile::WriteLine (unsigned char* _pValues, unsigned long _ulCount)
{
	if (0 == _ulCount)
		return true;
	*this << (unsigned int)_pValues [0];
	for (unsigned long i = 1; i < _ulCount; ++ i)
		*this << ',' << (unsigned int)_pValues [i];
	*this << endl;
	return true;
}


