#include "nlp_gslext.h"
#include <nlp_macros.h>


// ------------------------------------------------------------
CsvFile& operator<< (CsvFile& _rFile, gsl_vector* _pVector)
{
	if (0 == _pVector->size)
		return _rFile;

	_rFile << gsl_vector_get (_pVector, 0);
	for (size_t i = 1; i < _pVector->size; ++ i)
		_rFile << ',' << gsl_vector_get (_pVector, i);
	_rFile.flush ();
	return _rFile;
}


// ------------------------------------------------------------
CsvFile& operator<< (CsvFile& _rFile, gsl_matrix* _pMatrix)
{
	if ((0 == _pMatrix->size1) || (0 == _pMatrix->size2))
		return _rFile;

	_rFile << gsl_matrix_get (_pMatrix, 0, 0);
	for (size_t c = 1; c < _pMatrix->size2; ++c)
		_rFile << ',' << gsl_matrix_get (_pMatrix, 0, c);
	for (size_t r = 1; r < _pMatrix->size1; ++ r)
	{
		_rFile << endl;
		_rFile << gsl_matrix_get (_pMatrix, r, 0);
		for (size_t c = 1; c < _pMatrix->size2; ++c)
			_rFile << ',' << gsl_matrix_get (_pMatrix, r, c);
	}
	_rFile.flush ();
	return _rFile;
}


// ------------------------------------------------------------
CsvFile& operator>> (CsvFile& _rFile, gsl_vector** _ppVector)
{
	String sLine;
	if (false == ((File&)_rFile).ReadLine (sLine))
		return _rFile;
	char* pStart = (char*)(const char*)sLine;

	// find count of values on line ...
	unsigned long ulValues = 0;
	while ('\0' != *pStart)
	{
		while ((' ' == *pStart) || ('\t' == *pStart))
			++ pStart;

		++ ulValues;
		if (',' == *pStart)
		{
			++ pStart;
			continue;
		}

		pStart = strchr (pStart, ',');
		if (NULL == pStart)
			break;
		++ pStart;
	}

	// check & create new vector ...
	if (NULL != *_ppVector)
		gsl_vector_free (*_ppVector);
	*_ppVector = gsl_vector_alloc (ulValues);

	ulValues = 0;
	pStart = (char*)(const char*)sLine;
	while ('\0' != *pStart)
	{
		while ((' ' == *pStart) || ('\t' == *pStart))
			++ pStart;

		if (',' == *pStart)
		{
			gsl_vector_set (*_ppVector, ulValues, 0);
			++ ulValues;
			++ pStart;
			continue;
		}
		gsl_vector_set (*_ppVector, ulValues, atof (pStart));
		++ ulValues;
		pStart = strchr (pStart, ',');
		if (NULL == pStart)
			break;
		++ pStart;
	}

	return _rFile;
}


// ------------------------------------------------------------
typedef deque <double*>	DoublePtr_dq_t;

CsvFile& operator>> (CsvFile& _rFile, gsl_matrix** _ppMatrix)
{
	String sLine;
	if (false == ((File&)_rFile).ReadLine (sLine))
		return _rFile;
	char* pStart = (char*)(const char*)sLine;

	// find count of values on first line ...
	unsigned long ulColumns = 0;
	while ('\0' != *pStart)
	{
		while ((' ' == *pStart) || ('\t' == *pStart))
			++ pStart;

		++ ulColumns;
		if (',' == *pStart)
		{
			++ pStart;
			continue;
		}

		pStart = strchr (pStart, ',');
		if (NULL == pStart)
			break;
		++ pStart;
	}

	// load data to memory ...		
	DoublePtr_dq_t	dqValues;
	while ("" != sLine)
	{
		double* pValues = new double [ulColumns];
		dqValues.push_back (pValues);

		pStart = (char*)(const char*)sLine;
		while ('\0' != *pStart)
		{
			while ((' ' == *pStart) || ('\t' == *pStart))
				++ pStart;

			if (',' == *pStart)
			{
				*(pValues ++) = 0;
				++ pStart;
				continue;
			}
			*(pValues ++) = atof (pStart);
			pStart = strchr (pStart, ',');
			if (NULL == pStart)
				break;
			++ pStart;
		}
		if (false == ((File&)_rFile).ReadLine (sLine))
			break;
	}
	unsigned long ulRows = dqValues.size ();

	// check & create matrix ...	
	if (NULL != *_ppMatrix)
		gsl_matrix_free (*_ppMatrix);
	*_ppMatrix = gsl_matrix_alloc (ulRows, ulColumns);

	// set matrix values ...		
	unsigned long r = 0;
	ITERATE (DoublePtr_dq_t, dqValues, ite)
	{
		double* pRow = *ite;
		for (unsigned long c = 0; c < ulColumns; ++ c)
			gsl_matrix_set (*_ppMatrix, r, c, *(pRow ++));
		++ r;
	}

	return _rFile;
}



